"""Calculates loss per component."""
from typing import Any
import numpy as np
from matplotlib import pyplot as plt

import doplaydo.dodata as dd


def run(die_id: int, key: str = "components") -> dict[str, Any]:
    """Returns component loss in dB/component.

    Args:
        die_id: id of the die to analyze.
        key: key of the attribute to filter by.
        value: value of the attribute to filter by.
    """
    device_data_objects = dd.get_data_by_query(
        [
            dd.Die.id == die_id,
        ]
    )

    if not device_data_objects:
        raise ValueError(f"No device data found with die_id {die_id}, key {key!r}")

    powers = []
    components = []

    for device_data, df in device_data_objects:
        components.append(device_data.device.cell.attributes.get(key))
        powers.append(df.output_power.max())

    p = np.polyfit(components, powers, 1)
    component_loss = -p[0]

    fig = plt.figure()
    plt.plot(components, powers, "o")
    plt.plot(components, np.polyval(p, components), "r-", label="fit")

    plt.xlabel("Number of components")
    plt.ylabel("Power (dBm)")
    plt.title(f"loss = {-p[0]:.2f} dB/component")

    return dict(
        output={"component_loss": component_loss},
        summary_plot=fig,
        die_id=die_id,
    )


if __name__ == "__main__":
    d = run(10024)
    print(d["output"]["component_loss"])
